"""ASGI-Tools Utils."""

from functools import wraps
from inspect import iscoroutinefunction, isasyncgenfunction
from typing import Callable, Awaitable

from multidict import CIMultiDict

from .typing import ScopeHeaders


def is_awaitable(fn: Callable) -> bool:
    """Check than the given function is awaitable."""
    return iscoroutinefunction(fn) or isasyncgenfunction(fn)


def to_awaitable(fn: Callable) -> Callable[..., Awaitable]:
    """Convert the given function to a coroutine function if it isn't"""
    if is_awaitable(fn):
        return fn

    @wraps(fn)
    async def coro(*args, **kwargs):
        return fn(*args, **kwargs)

    return coro


def parse_headers(headers: ScopeHeaders) -> CIMultiDict:
    """Decode the given headers list."""
    return CIMultiDict([(n.decode('latin-1'), v.decode('latin-1')) for n, v in headers])



"""Work with multipart."""

from io import BytesIO
import typing as t
from cgi import parse_header
from tempfile import SpooledTemporaryFile
from urllib.parse import unquote_to_bytes

from multidict import MultiDict

from .multipart import QueryStringParser, MultipartParser, BaseParser
from .request import Request


async def read_formdata(request: Request, max_size: int, upload_to: t.Optional[t.Callable],
                        file_memory_limit: int = 1024 * 1024) -> MultiDict:
    """Read formdata from the given request."""
    if request.content_type == 'multipart/form-data':
        reader: FormReader = MultipartReader(request.charset, upload_to, file_memory_limit)
    else:
        reader = FormReader(request.charset)

    parser = reader.init_parser(request, max_size)
    async for chunk in request.stream():
        parser.write(chunk)

    parser.finalize()
    return reader.form


class FormReader:
    """Process querystring form data."""

    __slots__ = 'form', 'curname', 'curvalue', 'charset'

    def __init__(self, charset: str):
        self.charset = charset
        self.curname = bytearray()
        self.curvalue = bytearray()
        self.form: MultiDict = MultiDict()

    def init_parser(self, request: Request, max_size: int) -> BaseParser:
        return QueryStringParser({
            'field_name': self.on_field_name,
            'field_data': self.on_field_data,
            'field_end': self.on_field_end
        }, max_size=max_size)

    def on_field_name(self, data: bytes, start: int, end: int):
        self.curname += data[start:end]

    def on_field_data(self, data: bytes, start: int, end: int):
        self.curvalue += data[start:end]

    def on_field_end(self, data: bytes, start: int, end: int):
        self.form.add(
            unquote_plus(self.curname).decode(self.charset),
            unquote_plus(self.curvalue).decode(self.charset),
        )
        self.curname.clear()
        self.curvalue.clear()


class MultipartReader(FormReader):
    """Process multipart formdata."""

    __slots__ = ('form', 'curname', 'curvalue', 'charset', 'name',
                 'partdata', 'headers', 'upload_to', 'file_memory_limit')

    def __init__(self, charset: str, upload_to: t.Optional[t.Callable], file_memory_limit: int):
        super(MultipartReader, self).__init__(charset)
        self.name = ''
        self.headers: t.Dict[bytes, bytes] = {}
        self.partdata = BytesIO()
        self.upload_to = upload_to
        self.file_memory_limit = file_memory_limit

    def init_parser(self, request: Request, max_size: int) -> BaseParser:
        boundary = request.media.get('boundary', '')
        if not len(boundary):
            raise ValueError('Invalid content type boundary')

        return MultipartParser(request.media.get('boundary'), {
            'header_end': self.on_header_end,
            'header_field': self.on_header_field,
            'headers_finished': self.on_headers_finished,
            'header_value': self.on_header_value,
            'part_data': self.on_part_data,
            'part_end': self.on_part_end,
        }, max_size=max_size)

    def on_header_field(self, data: bytes, start: int, end: int):
        self.curname += data[start:end]

    def on_header_value(self, data: bytes, start: int, end: int):
        self.curvalue += data[start:end]

    def on_header_end(self, data: bytes, start: int, end: int):
        self.headers[bytes(self.curname.lower())] = bytes(self.curvalue)
        self.curname.clear()
        self.curvalue.clear()

    def on_headers_finished(self, data: bytes, start: int, end: int):
        _, options = parse_header(self.headers[b'content-disposition'].decode(self.charset))
        self.name = options['name']
        if 'filename' in options:
            upload_to = self.upload_to
            if upload_to is not None:
                filename = upload_to(options['filename'])
                self.partdata = f = open(filename, 'wb+')  # type: ignore

            else:
                self.partdata = f = SpooledTemporaryFile(self.file_memory_limit)  # type: ignore
                f._file.name = options['filename']  # type: ignore

            f.content_type = self.headers[b'content-type'].decode(self.charset)  # type: ignore

    def on_part_data(self, data: bytes, start: int, end: int):
        self.partdata.write(data[start:end])

    def on_part_end(self, data: bytes, start: int, end: int):
        field_data = self.partdata
        if isinstance(field_data, BytesIO):
            self.form.add(self.name, field_data.getvalue().decode(self.charset))

        else:
            field_data.seek(0)
            self.form.add(self.name, field_data)

        self.partdata = BytesIO()
        self.headers = {}


def unquote_plus(value: bytearray) -> bytes:
    value = value.replace(b'+', b' ')
    return unquote_to_bytes(bytes(value))

# pylama: ignore=D



"""ASGI-Tools Middlewares."""

import abc
import inspect
import typing as t
from functools import partial
from pathlib import Path

from http_router import Router

from . import ASGIError, asgi_logger
from .request import Request
from .response import parse_response, ResponseError, ResponseFile, Response, ResponseRedirect
from .typing import Scope, Receive, Send, ASGIApp


class BaseMiddeware(metaclass=abc.ABCMeta):
    """Base class for ASGI-Tools middlewares."""

    scopes: t.Set = {'http', 'websocket'}

    def __init__(self, app: ASGIApp = None) -> None:
        """Save ASGI App."""

        self.bind(app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """Handle ASGI call."""

        if scope['type'] in self.scopes:
            return await self.__process__(scope, receive, send)

        return await self.app(scope, receive, send)

    @abc.abstractmethod
    async def __process__(self, scope: Scope, receive: Receive, send: Send):
        """Do the middleware's logic."""

        raise NotImplementedError()

    @classmethod
    def setup(cls, **params) -> t.Callable:
        """Setup the middleware without an initialization."""
        return partial(cls, **params)

    def bind(self, app: ASGIApp = None):
        """Rebind the middleware to an ASGI application if it has been inited already."""
        self.app = app or ResponseError.NOT_FOUND()
        return self


class ResponseMiddleware(BaseMiddeware):
    """Automatically convert ASGI_ apps results into responses :class:`~asgi_tools.Response` and
    send them to server as ASGI_ messages.

    .. code-block:: python

        from asgi_tools import ResponseMiddleware, ResponseText, ResponseRedirect

        async def app(scope, receive, send):
            # ResponseMiddleware catches ResponseError, ResponseRedirect and convert the exceptions
            # into HTTP response
            if scope['path'] == '/user':
                raise ResponseRedirect('/login')

            # Return ResponseHTML
            if scope['method'] == 'GET':
                return '<b>HTML is here</b>'

            # Return ResponseJSON
            if scope['method'] == 'POST':
                return {'json': 'here'}

            # Return any response explicitly
            if scope['method'] == 'PUT':
                return ResponseText('response is here')

            # Short form to responses: (status_code, body) or (status_code, body, headers)
            return 405, 'Unknown method'

        app = ResponseMiddleware(app)

    The conversion rules:

    * :class:`Response` objects will be directly returned from the view
    * ``dict``, ``list``, ``int``, ``bool``, ``None`` results will be converted into :class:`ResponseJSON`
    * ``str``, ``bytes`` results will be converted into :class:`ResponseHTML`
    * ``tuple[int, Any, dict]`` will be converted into a :class:`Response` with ``int`` status code, ``dict`` will be used as headers, ``Any`` will be used to define the response's type

    .. code-block:: python

        from asgi_tools import ResponseMiddleware

        # The result will be converted into HTML 404 response with the 'Not Found' body
        async def app(request, receive, send):
            return 404, 'Not Found'

        app = ResponseMiddleware(app)

    You are able to raise :class:`ResponseError` from yours ASGI_ apps and it will be catched and returned as a response

    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """Handle ASGI call."""
        response = await self.__process__(scope, receive, send)
        if isinstance(response, Response):
            await response(scope, receive, send)

    async def __process__(self, scope: Scope, receive: Receive, send: Send):
        """Parse responses from callbacks."""

        try:
            response = await self.app(scope, receive, send)
            if response is None and scope['type'] == 'websocket':
                return

            # Prepare a response
            return parse_response(response)

        except (ResponseError, ResponseRedirect) as exc:
            return exc


class RequestMiddleware(BaseMiddeware):
    """Automatically create :class:`asgi_tools.Request` from the scope and pass it to ASGI_ apps.

    .. code-block:: python

        from asgi_tools import RequestMiddleware, Response

        async def app(request, receive, send):
            content = f"{ request.method } { request.url.path }"
            response = Response(content)
            await response(scope, receive, send)

        app = RequestMiddleware(app)

    """

    async def __process__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Replace scope with request object."""
        return await self.app(Request(scope, receive, send), receive, send)


class LifespanMiddleware(BaseMiddeware):
    """Manage ASGI_ Lifespan events.

    :param ignore_errors: Ignore errors from startup/shutdown handlers
    :param on_startup: the list of callables to run when the app is starting
    :param on_shutdown: the list of callables to run when the app is finishing

    .. code-block:: python

        from asgi_tools import LifespanMiddleware, Response

        async def app(scope, receive, send):
            response = Response('OK')
            await response(scope, receive, send)

        app = lifespan = LifespanMiddleware(app)

        @lifespan.on_startup
        async def start():
            print('The app is starting')

        @lifespan.on_shutdown
        async def start():
            print('The app is finishing')

    Lifespan middleware may be used as an async context manager for testing purposes

    .. code-block: python

        async def test_my_app():

            # ...

            # Registered startup/shutdown handlers will be called
            async with lifespan:
                # ... do something

    """

    scopes = {'lifespan'}

    def __init__(self, app: ASGIApp = None, ignore_errors: bool = False, logger=asgi_logger,
                 on_startup: t.Union[t.Callable, t.List[t.Callable]] = None,
                 on_shutdown: t.Union[t.Callable, t.List[t.Callable]] = None) -> None:
        """Prepare the middleware."""
        super(LifespanMiddleware, self).__init__(app)
        self.ignore_errors = ignore_errors
        self.logger = logger
        self.__startup__: t.List[t.Callable] = []
        self.__shutdown__: t.List[t.Callable] = []
        self.__register__(on_startup, self.__startup__)
        self.__register__(on_shutdown, self.__shutdown__)

    async def __process__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Manage lifespan cycle."""
        while True:
            message = await receive()
            if message['type'] == 'lifespan.startup':
                msg = await self.run('startup', send)
                await send(msg)

            elif message['type'] == 'lifespan.shutdown':
                msg = await self.run('shutdown', send)
                return await send(msg)

    def __register__(self, handlers: t.Union[t.Callable, t.List[t.Callable], None],
                     container: t.List[t.Callable]) -> None:
        """Register lifespan handlers."""
        if not handlers:
            return

        if not isinstance(handlers, list):
            handlers = [handlers]

        container += handlers

    async def __aenter__(self):
        """Use the lifespan middleware as a context manager."""
        await self.run('startup')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Use the lifespan middleware as a context manager."""
        await self.run('shutdown')

    async def run(self, event: str, send: Send = None):
        """Run startup/shutdown handlers."""
        assert event in {'startup', 'shutdown'}
        handlers = getattr(self, f"__{event}__")
        for handler in handlers:
            try:
                res = handler()
                if inspect.isawaitable(res):
                    await res
            except Exception as exc:
                if self.ignore_errors:
                    continue

                self.logger.exception(exc)
                self.logger.error(
                    f"{ event.title() } method '{ handler }' raises an exception. "
                    "Lifespan process failed.")

                return {'type': f'lifespan.{event}.failed', 'message': str(exc)}

        return {'type': f'lifespan.{event}.complete'}

    def on_startup(self, fn: t.Callable) -> None:
        """Add a function to startup."""
        self.__register__(fn, self.__startup__)

    def on_shutdown(self, fn: t.Callable) -> None:
        """Add a function to shutdown."""
        self.__register__(fn, self.__shutdown__)


class RouterMiddleware(BaseMiddeware):
    r"""Manage routing.

    .. code-block:: python

        from asgi_tools import RouterMiddleware, ResponseHTML, ResponseError

        async def default_app(scope, receive, send):
            response = ResponseError.NOT_FOUND()
            await response(scope, receive, send)

        app = router = RouterMiddleware(default_app)

        @router.route('/status', '/stat')
        async def status(scope, receive, send):
            response = ResponseHTML('STATUS OK')
            await response(scope, receive, send)

        # Bind methods
        # ------------
        @router.route('/only-post', methods=['POST'])
        async def only_post(scope, receive, send):
            response = ResponseHTML('POST OK')
            await response(scope, receive, send)

        # Regexp paths
        # ------------
        import re

        @router.route(re.compile(r'/\d+/?'))
        async def num(scope, receive, send):
            num = int(scope['path'].strip('/'))
            response = ResponseHTML(f'Number { num }')
            await response(scope, receive, send)

        # Dynamic paths
        # -------------

        @router.route('/hello/{name}')
        async def hello(scope, receive, send):
            name = scope['path_params']['name']
            response = ResponseHTML(f'Hello { name.title() }')
            await response(scope, receive, send)

        # Set regexp for params
        @router.route(r'/multiply/{first:\d+}/{second:\d+}')
        async def multiply(scope, receive, send):
            first, second = map(int, scope['path_params'].values())
            response = ResponseHTML(str(first * second))
            await response(scope, receive, send)

    Path parameters are made available in the request/scope, as the ``path_params`` dictionary.

    """

    def __init__(self, app: ASGIApp = None, router: Router = None) -> None:
        """Initialize HTTP router. """
        super(RouterMiddleware, self).__init__(app)
        self.router = router or Router()

    async def __process__(self, scope: Scope, receive: Receive, send: Send):
        """Get an app and process."""
        app, path_params = self.__dispatch__(scope)
        if not callable(app):
            app = self.app

        scope['path_params'] = path_params
        return await app(scope, receive, send)

    def __dispatch__(self, scope: Scope) -> t.Tuple[t.Optional[t.Any], t.Optional[t.Mapping]]:
        """Lookup for a callback."""
        try:
            match = self.router(scope.get("root_path", "") + scope["path"], scope['method'])
            return match.target, match.params

        except self.router.RouterError:
            return self.app, {}

    def route(self, *args, **kwargs):
        """Register a route."""
        return self.router.route(*args, **kwargs)


class StaticFilesMiddleware(BaseMiddeware):
    """Serve static files.

    :param url_prefix:  an URL prefix for static files
    :type url_prefix: str, "/static"
    :param folders: Paths to folders with static files
    :type folders: list[str]

    .. code-block:: python

        from asgi_tools import StaticFilesMiddleware, ResponseHTML

        async def app(scope, receive, send):
            response = ResponseHTML('OK)

        app = StaticFilesMiddleware(app, folders=['static'])

    """

    def __init__(self, app: ASGIApp = None, url_prefix: str = '/static',
                 folders: t.Union[str, t.List[str]] = None) -> None:
        """Initialize the middleware. """
        super(StaticFilesMiddleware, self).__init__(app)
        self.url_prefix = url_prefix
        folders = folders or []
        if isinstance(folders, str):
            folders = [folders]
        self.folders: t.List[Path] = [Path(folder) for folder in folders]

    async def __process__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Serve static files for self url prefix."""
        path = scope['path']
        url_prefix = self.url_prefix
        if not path.startswith(url_prefix):
            return await self.app(scope, receive, send)

        filename = path[len(url_prefix):].strip('/')
        for folder in self.folders:
            filepath = folder.joinpath(filename).resolve()
            try:
                response: t.Optional[Response] = ResponseFile(
                    filepath, headers_only=scope['method'] == 'HEAD')
                break

            except ASGIError:
                response = None

        response = response or ResponseError(status_code=404)
        await response(scope, receive, send)

# pylama: ignore=E501



"""ASGI-Tools includes a `asgi_tools.Request` class that gives you a nicer interface onto the
incoming request.
"""

from __future__ import annotations

import typing as t
from cgi import parse_header
from http import cookies

from multidict import MultiDict
from yarl import URL

from . import ASGIDecodeError, DEFAULT_CHARSET
from ._compat import json_loads
from .typing import Scope, Receive, Send, JSONType
from .utils import parse_headers, CIMultiDict


class Request(t.MutableMapping):
    """Represent a HTTP Request.

    :param scope: HTTP ASGI Scope
    :param receive: an asynchronous callable which lets the application
                    receive event messages from the client

    """

    _is_read: bool = False
    _url: t.Optional[URL] = None
    _body: t.Optional[bytes] = None
    _form: t.Optional[MultiDict] = None
    _headers: t.Optional[CIMultiDict] = None
    _media: t.Optional[t.Dict[str, str]] = None
    _cookies: t.Optional[t.Dict[str, str]] = None

    def __init__(self, scope: Scope, receive: Receive = None, send: Send = None) -> None:
        """Create a request based on the given scope."""
        self.scope = scope
        self.receive = receive
        self.send = send

    def __repr__(self):
        """Represent the request."""
        return f"<Request {self.content_type} {self.method} {self.url.path}>"

    def __getitem__(self, key: str) -> t.Any:
        """Proxy the method to the scope."""
        return self.scope[key]

    def __setitem__(self, key: str, value: t.Any) -> None:
        """Proxy the method to the scope."""
        self.scope[key] = value

    def __delitem__(self, key: str) -> None:
        """Proxy the method to the scope."""
        del self.scope[key]

    def __iter__(self) -> t.Iterator[str]:
        """Proxy the method to the scope."""
        return iter(self.scope)

    def __len__(self) -> int:
        """Proxy the method to the scope."""
        return len(self.scope)

    def __getattr__(self, name: str) -> t.Any:
        """Proxy the request's unknown attributes to scope."""
        return self.scope[name]

    def __copy__(self, **mutations) -> Request:
        """Copy the request to a new one."""
        return Request(dict(self.scope, **mutations), self.receive, self.send)

    @property
    def url(self) -> URL:
        """A lazy property that parses the current URL and returns :class:`yarl.URL` object.

        .. code-block:: python

            request = Request(scope)
            assert str(request.url) == '... the full http URL ..'
            assert request.url.scheme
            assert request.url.host
            assert request.url.query is not None
            assert request.url.query_string is not None

        See :py:mod:`yarl` documentation for further reference.

        """
        if self._url is None:
            scope = self.scope
            host = self.headers.get('host')
            if host is None and ('server' in scope):
                host, port = scope['server']
                if port:
                    host = f"{host}:{port}"

            self._url = URL.build(
                host=host,  # type: ignore
                scheme=scope.get('scheme', 'http'),
                encoded=True, path=f"{ scope.get('root_path', '') }{ scope['path'] }",
                query_string=scope['query_string'].decode("latin-1"),
            )

        return self._url

    @property
    def headers(self) -> CIMultiDict:
        """ A lazy property that parses the current scope's headers, decodes them as strings and
        returns case-insensitive multi-dict :py:class:`multidict.CIMultiDict`.

        .. code-block:: python

            request = Request(scope)

            assert request.headers['content-type']
            assert request.headers['authorization']

        See :py:mod:`multidict` documentation for futher reference.

        """
        if self._headers is None:
            self._headers = parse_headers(self.scope['headers'])
        return self._headers

    @property
    def cookies(self) -> t.Dict[str, str]:
        """A lazy property that parses the current scope's cookies and returns a dictionary.

        .. code-block:: python

            request = Request(scope)
            ses = request.cookies.get('session')

        """
        if self._cookies is None:
            self._cookies = {}
            cookie = self.headers.get('cookie')
            if cookie:
                for chunk in cookie.split(';'):
                    key, _, val = chunk.partition('=')
                    self._cookies[key.strip()] = cookies._unquote(val.strip())  # type: ignore

        return self._cookies

    @property
    def media(self) -> t.Dict[str, str]:
        """Prepare a media data for the request."""
        if self._media is None:
            content_type, opts = parse_header(self.headers.get('content-type', ''))
            self._media = dict(opts, content_type=content_type)

        return self._media

    @property
    def charset(self) -> str:
        """Get an encoding charset for the current scope."""
        return self.media.get('charset', DEFAULT_CHARSET)

    @property
    def query(self) -> MultiDict:
        """A lazy property that parse the current query string and returns it as a
        :py:class:`multidict.MultiDict`.

        """
        return self.url.query

    @property
    def content_type(self) -> str:
        """Get a content type for the current scope."""
        return self.media['content_type']

    async def stream(self) -> t.AsyncGenerator:
        """Stream the request's body.

        The method provides byte chunks without storing the entire body to memory.
        Any subsequent calls to :py:meth:`body`, :py:meth:`form`, :py:meth:`json`
        or :py:meth:`data` will raise an error.

        .. warning::
            You can only read stream once. Second call raises an error. Save a readed stream into a
            variable if you need.

        """
        if not self.receive:
            raise RuntimeError('Request doesnt have a receive coroutine')

        if self._is_read:
            raise RuntimeError('Stream has been read')

        self._is_read = True
        message = await self.receive()
        yield message.get('body', b'')
        while message.get('more_body'):
            message = await self.receive()
            yield message.get('body', b'')

    async def body(self) -> bytes:
        """Read and return the request's body as bytes.

        `body = await request.body()`
        """
        if self._body is None:
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)

            self._body = b"".join(chunks)

        return self._body

    async def text(self) -> str:
        """Read and return the request's body as a string.

        `text = await request.text()`
        """
        body = await self.body()
        try:
            return body.decode(self.charset or DEFAULT_CHARSET)
        except (LookupError, ValueError):
            raise ASGIDecodeError('Invalid Encoding')

    async def json(self) -> JSONType:
        """Read and return the request's body as a JSON.

        `json = await request.json()`

        """
        try:
            return json_loads(await self.body())
        except (LookupError, ValueError):
            raise ASGIDecodeError('Invalid JSON')

    async def form(self, max_size: int = 0, upload_to: t.Callable = None,
                   file_memory_limit: int = 1024 * 1024) -> MultiDict:
        """Read and return the request's multipart formdata as a multidict.

        The method reads the request's stream stright into memory formdata.
        Any subsequent calls to :py:meth:`body`, :py:meth:`json` will raise an error.

        `formdata = await request.form()`

        """
        from .forms import read_formdata

        if self._form is None:
            try:
                self._form = await read_formdata(self, max_size, upload_to, file_memory_limit)
            except (LookupError, ValueError):
                raise ASGIDecodeError('Invalid Encoding')

        return self._form

    async def data(self, raise_errors: bool = False) -> t.Union[str, bytes, JSONType, MultiDict]:
        """The method checks Content-Type Header and parse the request's data automatically.

        `data = await request.data()`

        If `raise_errors` is false (by default) and the given data is invalid (ex. invalid json)
        the request's body would be returned.

        Returns data from :py:meth:`json` for `application/json`, :py:meth:`form` for
        `application/x-www-form-urlencoded`,  `multipart/form-data` and :py:meth:`text` otherwise.
        """
        try:
            if self.content_type in {'application/x-www-form-urlencoded', 'multipart/form-data'}:
                return await self.form()

            if self.content_type == 'application/json':
                return await self.json()

            return await self.text()

        except ASGIDecodeError:
            if raise_errors:
                raise
            return await self.body()



"""The code is based on a great work of Andrew Dunham
(https://github.com/andrew-d/python-multipart) and has been changed to improve speed.

The original code is licensed by Apache2 license.
"""
import typing as t

# Flags for the multipart parser.
FLAG_PART_BOUNDARY              = 1
FLAG_LAST_BOUNDARY              = 2

# Get constants.  Since iterating over a str on Python 2 gives you a 1-length
# string, but iterating over a bytes object on Python 3 gives you an integer,
# we need to save these constants.
AMPERSAND = b'&'[0]
COLON = b':'[0]
CR = b'\r'[0]
EQUAL = b'='[0]
HYPHEN = b'-'[0]
LF = b'\n'[0]
NULL = b'\x00'[0]
SEMICOLON = b';'[0]
SPACE = b' '[0]


class BaseParser:
    """This class is the base class for all parsers.  It contains the logic for
    calling and adding callbacks.

    A callback can be one of two different forms.  "Notification callbacks" are
    callbacks that are called when something happens - for example, when a new
    part of a multipart message is encountered by the parser.  "Data callbacks"
    are called when we get some sort of data - for example, part of the body of
    a multipart chunk.  Notification callbacks are called with no parameters,
    whereas data callbacks are called with three, as follows::

        data_callback(data, start, end)

    The "data" parameter is a bytestring. "start" and "end" are integer indexes into the "data"
    string that represent the data of interest.  Thus, in a data callback, the slice
    `data[start:end]` represents the data that the callback is "interested in".  The callback is
    not passed a copy of the data, since copying severely hurts performance.

    """

    def __init__(self, callbacks: t.Dict):
        self.callbacks = callbacks or {}

    def callback(self, name: str, data: bytes = None, start: int = 0, end: int = 0):
        if data is not None and start == end:
            return

        try:
            func = self.callbacks[name]
            func(data, start, end)
        except KeyError:
            pass

    def write(self, data: bytes):
        pass

    def finalize(self):
        pass


STATE_BEFORE_FIELD = 0
STATE_FIELD_NAME = 1
STATE_FIELD_DATA = 2


class QueryStringParser(BaseParser):
    """this is a streaming querystring parser.  it will consume data, and call
    the callbacks given when it has data.

    .. list-table::
       :widths: 15 10 30
       :header-rows: 1

       * - callback name
         - parameters
         - description
       * - field_start
         - none
         - called when a new field is encountered.
       * - field_name
         - data, start, end
         - called when a portion of a field's name is encountered.
       * - field_data
         - data, start, end
         - called when a portion of a field's data is encountered.
       * - field_end
         - none
         - called when the end of a field is encountered.
       * - end
         - none
         - called when the parser is finished parsing all data.

    :param callbacks: a dictionary of callbacks.  see the documentation for
                      :class:`baseparser`.

    :param max_size: the maximum size of body to parse.  defaults to 0
    """

    def __init__(self, callbacks: t.Dict, max_size: int = 0):
        self.callbacks = callbacks
        self.cursize = 0
        self.max_size = max_size

        self.state = STATE_BEFORE_FIELD

    def write(self, data: bytes):
        data_len = prune_data(len(data), self.cursize, self.max_size)

        idx = 0
        state = self.state

        while idx < data_len:
            ch = data[idx]
            if state == STATE_BEFORE_FIELD:

                if not (ch == AMPERSAND or ch == SEMICOLON):
                    self.callback('field_start')
                    idx -= 1
                    state = STATE_FIELD_NAME

            elif state == STATE_FIELD_NAME:
                sep_pos = data.find(AMPERSAND, idx)
                if sep_pos == -1:
                    sep_pos = data.find(SEMICOLON, idx)

                if sep_pos != -1:
                    equals_pos = data.find(EQUAL, idx, sep_pos)
                else:
                    equals_pos = data.find(EQUAL, idx)

                if equals_pos != -1:
                    self.callback('field_name', data, idx, equals_pos)
                    idx = equals_pos
                    state = STATE_FIELD_DATA

                else:
                    if sep_pos == -1:
                        self.callback('field_name', data, idx, data_len)
                        idx = data_len

                    else:
                        self.callback('field_name', data, idx, sep_pos)
                        self.callback('field_end')
                        idx = sep_pos - 1
                        state = STATE_BEFORE_FIELD

            elif state == STATE_FIELD_DATA:
                sep_pos = data.find(AMPERSAND, idx)
                if sep_pos == -1:
                    sep_pos = data.find(SEMICOLON, idx)

                if sep_pos == -1:
                    self.callback('field_data', data, idx, data_len)
                    idx = data_len

                else:
                    self.callback('field_data', data, idx, sep_pos)
                    self.callback('field_end')

                    idx = sep_pos - 1
                    state = STATE_BEFORE_FIELD

            else:
                raise ValueError(f"Reached an unknown state {state} at {idx}")

            idx += 1

        self.state = state
        self.cursize += data_len

    def finalize(self):
        """Finalize this parser, which signals to that we are finished parsing,
        if we're still in the middle of a field, an on_field_end callback, and
        then the on_end callback.
        """
        # If we're currently in the middle of a field, we finish it.
        if self.state == STATE_FIELD_DATA:
            self.callback('field_end')
        self.callback('end')


STATE_START                     = 0
STATE_START_BOUNDARY            = 1
STATE_HEADER_FIELD_START        = 2
STATE_HEADER_FIELD              = 3
STATE_HEADER_VALUE_START        = 4
STATE_HEADER_VALUE              = 5
STATE_HEADER_VALUE_ALMOST_DONE  = 6
STATE_HEADERS_ALMOST_DONE       = 7
STATE_PART_DATA_START           = 8
STATE_PART_DATA                 = 9
STATE_PART_DATA_END             = 10
STATE_END                       = 11


class MultipartParser(BaseParser):
    """This class is a streaming multipart/form-data parser.

    .. list-table::
       :widths: 15 10 30
       :header-rows: 1

       * - Callback Name
         - Parameters
         - Description
       * - part_begin
         - None
         - Called when a new part of the multipart message is encountered.
       * - part_data
         - data, start, end
         - Called when a portion of a part's data is encountered.
       * - part_end
         - None
         - Called when the end of a part is reached.
       * - header_begin
         - None
         - Called when we've found a new header in a part of a multipart
           message
       * - header_field
         - data, start, end
         - Called each time an additional portion of a header is read (i.e. the
           part of the header that is before the colon; the "Foo" in
           "Foo: Bar").
       * - header_value
         - data, start, end
         - Called when we get data for a header.
       * - header_end
         - None
         - Called when the current header is finished - i.e. we've reached the
           newline at the end of the header.
       * - headers_finished
         - None
         - Called when all headers are finished, and before the part data
           starts.
       * - end
         - None
         - Called when the parser is finished parsing all data.


    :param boundary: The multipart boundary.  This is required, and must match
                     what is given in the HTTP request - usually in the
                     Content-Type header.

    :param callbacks: A dictionary of callbacks.  See the documentation for
                      :class:`BaseParser`.

    :param max_size: The maximum size of body to parse.  Defaults to 0
    """

    def __init__(self, boundary, callbacks={}, max_size=0):
        self.callbacks = callbacks
        self.cursize = 0
        self.max_size = max_size
        self.state = STATE_START
        self.index = self.flags = 0

        self.header_field_pos = -1
        self.header_value_pos = -1
        self.part_data_pos = -1

        if isinstance(boundary, str):
            boundary = boundary.encode('latin-1')

        self.boundary = b'\r\n--' + boundary

        # Get a set of characters that belong to our boundary.
        self.boundary_chars = frozenset(self.boundary)

        # We also create a lookbehind list.
        # Note: the +8 is since we can have, at maximum, "\r\n--" + boundary +
        # "--\r\n" at the final boundary, and the length of '\r\n--' and
        # '--\r\n' is 8 bytes.
        self.lookbehind = [NULL for x in range(len(boundary) + 8)]

    def write(self, data):  # noqa
        data_len = prune_data(len(data), self.cursize, self.max_size)

        idx = 0
        index = self.index
        state = self.state
        flags = self.flags
        boundary = self.boundary
        boundary_len = len(boundary)

        while idx < data_len:
            ch = data[idx]

            if state == STATE_START_BOUNDARY:
                # Check to ensure that the last 2 characters in our boundary
                # are CRLF.
                if index == len(boundary) - 2:
                    if ch != CR:
                        raise ValueError(f"Did not find \\r at end of boundary ({idx})")
                    index += 1

                elif index == len(boundary) - 2 + 1:
                    if ch != LF:
                        raise ValueError(f"Did not find \\n at end of boundary ({idx})")

                    state = STATE_HEADER_FIELD_START
                    self.callback('part_begin')

                # Check to ensure our boundary matches
                elif ch == boundary[index + 2]:
                    # Increment index into boundary and continue.
                    index += 1

                else:
                    raise ValueError(f"Did not find boundary character {chr(ch)} at index {idx}")

            elif state == STATE_HEADER_FIELD_START:
                # Mark the start of a header field here, reset the index, and
                # continue parsing our header field.
                index = 0
                self.header_field_pos = idx
                idx -= 1
                state = STATE_HEADER_FIELD

            elif state == STATE_HEADER_FIELD:
                # If we've reached a CR at the beginning of a header, it means
                # that we've reached the second of 2 newlines, and so there are
                # no more headers to parse.
                if ch == CR:
                    self.header_field_pos = -1
                    state = STATE_HEADERS_ALMOST_DONE
                    idx += 1
                    continue

                index += 1

                # If we've reached a colon, we're done with this header.
                if ch == COLON:
                    # A 0-length header is an error.
                    if index == 1:
                        raise ValueError(f"Found 0-length header at {idx}")

                    # Call our callback with the header field.
                    if self.header_field_pos != -1:
                        self.callback('header_field', data, self.header_field_pos, idx)
                        self.header_field_pos = -1

                    # Move to parsing the header value.
                    state = STATE_HEADER_VALUE_START

            elif state == STATE_HEADER_VALUE_START:
                # Skip leading spaces.
                if ch != SPACE:
                    # Mark the start of the header value.
                    self.header_value_pos = idx
                    idx -= 1
                    # Move to the header-value state, reprocessing this character.
                    state = STATE_HEADER_VALUE

            elif state == STATE_HEADER_VALUE:
                # If we've got a CR, we're nearly done our headers.  Otherwise,
                # we do nothing and just move past this character.
                if ch == CR:
                    if self.header_value_pos != -1:
                        self.callback('header_value', data, self.header_value_pos, idx)
                        self.header_value_pos = -1

                    self.callback('header_end')
                    state = STATE_HEADER_VALUE_ALMOST_DONE

            elif state == STATE_HEADER_VALUE_ALMOST_DONE:
                # The last character should be a LF.  If not, it's an error.
                if ch != LF:
                    raise ValueError(f"Did not find \\n at end of header (found {chr(ch)})")

                # Move back to the start of another header.  Note that if that
                # state detects ANOTHER newline, it'll trigger the end of our
                # headers.
                state = STATE_HEADER_FIELD_START

            elif state == STATE_HEADERS_ALMOST_DONE:
                # We're almost done our headers.  This is reached when we parse
                # a CR at the beginning of a header, so our next character
                # should be a LF, or it's an error.
                if ch != LF:
                    raise ValueError(f"Did not find \\n at end of headers (found {chr(ch)})")

                self.callback('headers_finished')
                # Mark the start of our part data.
                self.part_data_pos = idx + 1
                state = STATE_PART_DATA

            elif state == STATE_PART_DATA:
                # We're processing our part data right now.  During this, we
                # need to efficiently search for our boundary, since any data
                # on any number of lines can be a part of the current data.
                # We use the Boyer-Moore-Horspool algorithm to efficiently
                # search through the remainder of the buffer looking for our
                # boundary.

                # Save the current value of our index.  We use this in case we
                # find part of a boundary, but it doesn't match fully.
                prev_index = index

                # Set up variables.
                boundary_end = boundary_len - 1
                boundary_chars = self.boundary_chars

                # If our index is 0, we're starting a new part, so start our
                # search.
                if index == 0:
                    # Search forward until we either hit the end of our buffer,
                    # or reach a character that's in our boundary.
                    idx += boundary_end
                    while idx < data_len - 1 and data[idx] not in boundary_chars:
                        idx += boundary_len

                    # Reset i back the length of our boundary, which is the
                    # earliest possible location that could be our match (i.e.
                    # if we've just broken out of our loop since we saw the
                    # last character in our boundary)
                    idx -= boundary_end
                    ch = data[idx]

                # Now, we have a couple of cases here.  If our index is before
                # the end of the boundary...
                if index < boundary_len:
                    # If the character matches...
                    if boundary[index] == ch:
                        # If we found a match for our boundary, we send the
                        # existing data.
                        if index == 0 and self.part_data_pos != -1:
                            self.callback('part_data', data, self.part_data_pos, idx)
                            self.part_data_pos = -1

                        # The current character matches, so continue!
                        index += 1
                    else:
                        index = 0

                # Our index is equal to the length of our boundary!
                elif index == boundary_len:
                    # First we increment it.
                    index += 1

                    # Now, if we've reached a newline, we need to set this as
                    # the potential end of our boundary.
                    if ch == CR:
                        flags |= FLAG_PART_BOUNDARY

                    # Otherwise, if this is a hyphen, we might be at the last
                    # of all boundaries.
                    elif ch == HYPHEN:
                        flags |= FLAG_LAST_BOUNDARY

                    # Otherwise, we reset our index, since this isn't either a
                    # newline or a hyphen.
                    else:
                        index = 0

                # Our index is right after the part boundary, which should be
                # a LF.
                elif index == boundary_len + 1:
                    # If we're at a part boundary (i.e. we've seen a CR
                    # character already)...
                    if flags & FLAG_PART_BOUNDARY:
                        # We need a LF character next.
                        if ch == LF:
                            # Unset the part boundary flag.
                            flags &= (~FLAG_PART_BOUNDARY)

                            # Callback indicating that we've reached the end of
                            # a part, and are starting a new one.
                            self.callback('part_end')
                            self.callback('part_begin')

                            # Move to parsing new headers.
                            index = 0
                            state = STATE_HEADER_FIELD_START
                            idx += 1
                            continue

                        # We didn't find an LF character, so no match.  Reset
                        # our index and clear our flag.
                        index = 0
                        flags &= (~FLAG_PART_BOUNDARY)

                    # Otherwise, if we're at the last boundary (i.e. we've
                    # seen a hyphen already)...
                    elif flags & FLAG_LAST_BOUNDARY:
                        # We need a second hyphen here.
                        if ch == HYPHEN:
                            # Callback to end the current part, and then the
                            # message.
                            self.callback('part_end')
                            self.callback('end')
                            state = STATE_END
                        else:
                            # No match, so reset index.
                            index = 0

                # If we have an index, we need to keep this byte for later, in
                # case we can't match the full boundary.
                if index > 0:
                    self.lookbehind[index - 1] = ch

                # Otherwise, our index is 0.  If the previous index is not, it
                # means we reset something, and we need to take the data we
                # thought was part of our boundary and send it along as actual
                # data.
                elif prev_index > 0:
                    # Callback to write the saved data.
                    lb_data = bytes(self.lookbehind)
                    self.callback('part_data', lb_data, 0, prev_index)

                    # Overwrite our previous index.
                    prev_index = 0

                    # Re-set our mark for part data.
                    self.part_data_pos = idx

                    # Re-consider the current character, since this could be
                    # the start of the boundary itself.
                    idx -= 1

            elif state == STATE_START:
                # Skip leading newlines
                if not (ch == CR or ch == LF):

                    # Move to the next state, but decrement i so that we re-process
                    # this character.
                    idx -= 1
                    state = STATE_START_BOUNDARY

            elif state == STATE_END:
                # Do nothing and just consume a byte in the end state.
                pass

            else:
                raise ValueError(f"Reached an unknown state {state} at {idx}")

            # Move to the next byte.
            idx += 1

        if self.header_field_pos != -1:
            self.callback('header_field', data, self.header_field_pos, data_len)
            self.header_field_pos = 0

        if self.header_value_pos != -1:
            self.callback('header_value', data, self.header_value_pos, data_len)
            self.header_value_pos = 0

        if self.part_data_pos != -1:
            self.callback('part_data', data, self.part_data_pos, data_len)
            self.part_data_pos = 0

        self.index = index
        self.state = state
        self.flags = flags
        self.cursize += data_len


def prune_data(data_len: int, cursize: int, max_size: int) -> int:
    if max_size and (cursize + data_len) > max_size:
        data_len = max_size - cursize

    return data_len


#  pylama:ignore=D,E221



"""ASGI Types."""

import typing as t


Message = t.MutableMapping[str, t.Any]
Send = t.Callable[[Message], t.Awaitable[t.Any]]
Receive = t.Callable[[], t.Awaitable[Message]]
Scope = t.MutableMapping[str, t.Any]
ScopeHeaders = t.List[t.Tuple[bytes, bytes]]
JSONType = t.Union[str, int, float, bool, None, t.Dict[str, t.Any], t.List[t.Any]]
ASGIApp = t.Callable[[Scope, Receive, Send], t.Awaitable]

ResponseContent = t.Union[str, bytes]
F = t.TypeVar('F', bound=t.Callable)
